@@ -22,6 +22,7 @@
 
 #include <json-glib/json-glib.h>
 
+#include <stdint.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -30,6 +31,13 @@
 #include <stdio.h>
 #include <sys/vfs.h>
 
+#ifdef HAVE_LIBSYSTEMD
+#include <systemd/sd-login.h>
+#include "sd-escape.h"
+#endif
+
+#include <gio/gio.h>
+#include <gio/gunixoutputstream.h>
 #include <gio/gdesktopappinfo.h>
 
 #include "xdp-utils.h"
@@ -106,13 +114,6 @@ xdp_mkstempat (int    dir_fd,
   return -1;
 }
 
-typedef enum
-{
-  XDP_APP_INFO_KIND_HOST = 0,
-  XDP_APP_INFO_KIND_FLATPAK = 1,
-  XDP_APP_INFO_KIND_SNAP    = 2,
-} XdpAppInfoKind;
-
 struct _XdpAppInfo {
   volatile gint ref_count;
   char *id;
@@ -124,7 +125,7 @@ struct _XdpAppInfo {
         {
           GKeyFile *keyfile;
 	   /* pid namespace mapping */
-          GMutex *pidns_lock;
+          GMutex pidns_lock;
           ino_t   pidns_id;
         } flatpak;
       struct
@@ -143,11 +144,81 @@ xdp_app_info_new (XdpAppInfoKind kind)
   return app_info;
 }
 
+char *
+_xdp_parse_app_id_from_unit_name (const char *unit)
+{
+  g_autoptr(GRegex) regex = NULL;
+  g_autoptr(GMatchInfo) match = NULL;
+  g_autoptr(GError) error = NULL;
+  g_autofree char *app_id = NULL;
+
+  g_assert (g_str_has_prefix (unit, "app-"));
+
+  /*
+   * From https://systemd.io/DESKTOP_ENVIRONMENTS/ the format is one of:
+   * app[-<launcher>]-<ApplicationID>-<RANDOM>.scope
+   * app[-<launcher>]-<ApplicationID>-autostart.service -> no longer true since systemd v248
+   * app[-<launcher>]-<ApplicationID>[@<RANDOM>].service
+   * app[-<launcher>]-<ApplicationID>-<RANDOM>.slice
+   */
+  regex = g_regex_new ("^app-(?:[[:alnum:]]+\\-)?(.+?\\..+?\\..+?)"
+                       "(?:[@\\-][0-9]*)?(?:-autostart)?"
+                       "(?:\\.scope|\\.service|\\.slice)$",
+                       0, 0, &error);
+  g_assert (error == NULL);
+  if (g_regex_match (regex, unit, 0, &match))
+    {
+      const char *escaped_app_id;
+      /* Unescape the unit name which may have \x hex codes in it, e.g.
+       * "app-gnome-org.gnome.Evolution\x2dalarm\x2dnotify-2437.scope"
+       */
+      escaped_app_id = g_match_info_fetch (match, 1);
+      if (cunescape (escaped_app_id, UNESCAPE_RELAX, &app_id) < 0)
+        app_id = g_strdup ("");
+    }
+  else
+    {
+      app_id = g_strdup ("");
+    }
+
+  return g_steal_pointer (&app_id);
+}
+
+void
+set_appid_from_pid (XdpAppInfo *app_info, pid_t pid)
+{
+#ifdef HAVE_LIBSYSTEMD
+  g_autofree char *unit = NULL;
+  int res;
+
+  g_return_if_fail (app_info->id == NULL);
+
+  res = sd_pid_get_user_unit (pid, &unit);
+  /*
+   * The session might not be managed by systemd or there could be an error
+   * fetching our own systemd units or the unit might not be started by the
+   * desktop environment (e.g. it's a script run from terminal).
+   */
+  if (res == -ENODATA || res < 0 || !unit || !g_str_has_prefix (unit, "app-"))
+    {
+      app_info->id = g_strdup ("");
+      return;
+    }
+
+  app_info->id = _xdp_parse_app_id_from_unit_name (unit);
+  g_debug ("Assigning app ID \"%s\" to pid %ld which has unit \"%s\"",
+           app_info->id, (long) pid, unit);
+
+#else
+  app_info->id = g_strdup ("");
+#endif /* HAVE_LIBSYSTEMD */
+}
+
 static XdpAppInfo *
-xdp_app_info_new_host (void)
+xdp_app_info_new_host (pid_t pid)
 {
   XdpAppInfo *app_info = xdp_app_info_new (XDP_APP_INFO_KIND_HOST);
-  app_info->id = g_strdup ("");
+  set_appid_from_pid (app_info, pid);
   return app_info;
 }
 
@@ -200,6 +271,14 @@ xdp_app_info_get_id (XdpAppInfo *app_info)
   return app_info->id;
 }
 
+XdpAppInfoKind
+xdp_app_info_get_kind (XdpAppInfo  *app_info)
+{
+  g_return_val_if_fail (app_info != NULL, -1);
+
+  return app_info->kind;
+}
+
 GAppInfo *
 xdp_app_info_load_app_info (XdpAppInfo *app_info)
 {
@@ -232,20 +311,52 @@ xdp_app_info_load_app_info (XdpAppInfo *app_info)
   return G_APP_INFO (g_desktop_app_info_new (desktop_id));
 }
 
+static gboolean
+needs_quoting (const char *arg)
+{
+  while (*arg != 0)
+    {
+      char c = *arg;
+      if (!g_ascii_isalnum (c) &&
+          !(c == '-' || c == '/' || c == '~' ||
+            c == ':' || c == '.' || c == '_' ||
+            c == '=' || c == '@'))
+        return TRUE;
+      arg++;
+    }
+  return FALSE;
+}
+
+static char *
+maybe_quote (const char *arg,
+             gboolean quote_escape)
+{
+  if (!quote_escape || !needs_quoting (arg))
+    return g_strdup (arg);
+  else
+    return g_shell_quote (arg);
+}
+
 char **
 xdp_app_info_rewrite_commandline (XdpAppInfo *app_info,
-                                  const char * const *commandline)
+                                  const char * const *commandline,
+                                  gboolean quote_escape)
 {
+  g_autoptr(GPtrArray) args = NULL;
+
   g_return_val_if_fail (app_info != NULL, NULL);
 
   if (app_info->kind == XDP_APP_INFO_KIND_HOST)
     {
-      return g_strdupv ((char **)commandline);
+      int i;
+      args = g_ptr_array_new_with_free_func (g_free);
+      for (i = 0; commandline && commandline[i]; i++)
+        g_ptr_array_add (args, maybe_quote (commandline[i], quote_escape));
+      g_ptr_array_add (args, NULL);
+      return (char **)g_ptr_array_free (g_steal_pointer (&args), FALSE);
     }
   else if (app_info->kind == XDP_APP_INFO_KIND_FLATPAK)
     {
-      g_autoptr(GPtrArray) args = NULL;
-
       args = g_ptr_array_new_with_free_func (g_free);
 
       g_ptr_array_add (args, g_strdup ("flatpak"));
@@ -253,14 +364,21 @@ xdp_app_info_rewrite_commandline (XdpAppInfo *app_info,
       if (commandline && commandline[0])
         {
           int i;
+          g_autofree char *quoted_command = NULL;
 
-          g_ptr_array_add (args, g_strdup_printf ("--command=%s", commandline[0]));
-          g_ptr_array_add (args, g_strdup (app_info->id));
+          quoted_command = maybe_quote (commandline[0], quote_escape);
+
+          g_ptr_array_add (args, g_strdup_printf ("--command=%s", quoted_command));
+
+          /* Always quote the app ID to make rewriting the file simpler in case
+           * the app is renamed.
+           */
+          g_ptr_array_add (args, g_shell_quote (app_info->id));
           for (i = 1; commandline[i]; i++)
-            g_ptr_array_add (args, g_strdup (commandline[i]));
+            g_ptr_array_add (args, maybe_quote (commandline[i], quote_escape));
         }
       else
-        g_ptr_array_add (args, g_strdup (app_info->id));
+        g_ptr_array_add (args, g_shell_quote (app_info->id));
       g_ptr_array_add (args, NULL);
 
       return (char **)g_ptr_array_free (g_steal_pointer (&args), FALSE);
@@ -269,6 +387,56 @@ xdp_app_info_rewrite_commandline (XdpAppInfo *app_info,
     return NULL;
 }
 
+char *
+xdp_app_info_get_tryexec_path (XdpAppInfo *app_info)
+{
+  g_return_val_if_fail (app_info != NULL, NULL);
+
+  if (app_info->kind == XDP_APP_INFO_KIND_FLATPAK)
+    {
+      g_autofree char *original_app_path = NULL;
+      g_autofree char *tryexec_path = NULL;
+      g_autofree char *app_slash = NULL;
+      g_autofree char *app_path = NULL;
+      char *app_slash_pointer;
+      char *path;
+
+      original_app_path = g_key_file_get_string (app_info->u.flatpak.keyfile,
+                                                 FLATPAK_METADATA_GROUP_INSTANCE,
+                                                 FLATPAK_METADATA_KEY_ORIGINAL_APP_PATH, NULL);
+      app_path = g_key_file_get_string (app_info->u.flatpak.keyfile,
+                                        FLATPAK_METADATA_GROUP_INSTANCE,
+                                        FLATPAK_METADATA_KEY_APP_PATH, NULL);
+      path = original_app_path ? original_app_path : app_path;
+
+      if (path == NULL || *path == '\0')
+        return NULL;
+
+      app_slash = g_strconcat ("app/", app_info->id, NULL);
+
+      app_slash_pointer = strstr (path, app_slash);
+      if (app_slash_pointer == NULL)
+        return NULL;
+
+      /* Terminate path after the flatpak installation path such as .local/share/flatpak/ */
+      *app_slash_pointer = '\0';
+
+      /* Find the path to the wrapper script exported by Flatpak, which can be
+       * used in a desktop file's TryExec=
+       */
+      tryexec_path = g_strconcat (path, "exports/bin/", app_info->id, NULL);
+      if (access (tryexec_path, X_OK) != 0)
+        {
+          g_debug ("Wrapper script unexpectedly not executable or nonexistent: %s", tryexec_path);
+          return NULL;
+        }
+
+      return g_steal_pointer (&tryexec_path);
+    }
+  else
+    return NULL;
+}
+
 char *
 xdp_app_info_get_instance (XdpAppInfo *app_info)
 {
@@ -331,6 +499,20 @@ xdp_app_info_remap_path (XdpAppInfo *app_info,
         return g_build_filename ("/usr", path + strlen ("/run/host/usr/"), NULL);
       else if (g_str_has_prefix (path, "/run/host/etc/"))
         return g_build_filename ("/etc", path + strlen ("/run/host/etc/"), NULL);
+      else if (g_str_has_prefix (path, "/run/flatpak/app/"))
+        return g_build_filename (g_get_user_runtime_dir (), "app",
+                                 path + strlen ("/run/flatpak/app/"), NULL);
+      else if (g_str_has_prefix (path, "/run/flatpak/doc/"))
+        return g_build_filename (g_get_user_runtime_dir (), "doc",
+                                 path + strlen ("/run/flatpak/doc/"), NULL);
+      else if (g_str_has_prefix (path, "/var/config/"))
+        return g_build_filename (g_get_home_dir (), ".var", "app",
+                                 app_info->id, "config",
+                                 path + strlen ("/var/config/"), NULL);
+      else if (g_str_has_prefix (path, "/var/data/"))
+        return g_build_filename (g_get_home_dir (), ".var", "app",
+                                 app_info->id, "data",
+                                 path + strlen ("/var/data/"), NULL);
     }
 
   return g_strdup (path);
@@ -472,7 +654,7 @@ parse_app_info_from_flatpak_info (int pid, GError **error)
     group = "Runtime";
 
   id = g_key_file_get_string (metadata, group, "name", error);
-  if (id == NULL)
+  if (id == NULL || !xdp_is_valid_app_id (id))
     {
       close (info_fd);
       return NULL;
@@ -511,9 +693,9 @@ _xdp_parse_cgroup_file (FILE *f, gboolean *is_snap)
 
       /* Only consider the freezer, systemd group or unified cgroup
        * hierarchies */
-      if ((!strcmp (controller, "freezer:") != 0 ||
-           !strcmp (controller, "name=systemd:") != 0 ||
-           !strcmp (controller, ":") != 0) &&
+      if ((strcmp (controller, "freezer:") == 0 ||
+           strcmp (controller, "name=systemd:") == 0 ||
+           strcmp (controller, ":") == 0) &&
           strstr (cgroup, "/snap.") != NULL)
         {
           *is_snap = TRUE;
@@ -641,7 +823,7 @@ xdp_get_app_info_from_pid (pid_t pid,
     }
 
   if (app_info == NULL)
-    app_info = xdp_app_info_new_host ();
+    app_info = xdp_app_info_new_host (pid);
 
   return g_steal_pointer (&app_info);
 }
@@ -854,14 +1036,21 @@ xdg_desktop_portal_error_quark (void)
 
 static char *
 verify_proc_self_fd (XdpAppInfo *app_info,
-                     const char *proc_path)
+                     const char *proc_path,
+                     GError **error)
 {
   char path_buffer[PATH_MAX + 1];
   ssize_t symlink_size;
+  int saved_errno;
 
   symlink_size = readlink (proc_path, path_buffer, PATH_MAX);
   if (symlink_size < 0)
-    return NULL;
+    {
+      saved_errno = errno;
+      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (saved_errno),
+                   "readlink %s: %s", proc_path, g_strerror (saved_errno));
+      return NULL;
+    }
 
   path_buffer[symlink_size] = 0;
 
@@ -869,16 +1058,24 @@ verify_proc_self_fd (XdpAppInfo *app_info,
      don't, such as socket:[27345] or anon_inode:[eventfd].
      We don't support any of these */
   if (path_buffer[0] != '/')
-    return NULL;
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_FILENAME,
+                   "Not a regular file or directory: %s", path_buffer);
+      return NULL;
+    }
 
   /* File descriptors to actually deleted files have " (deleted)"
      appended to them. This also happens to some fake fd types
      like shmem which are "/<name> (deleted)". All such
-     files are considered invalid. Unfortunatelly this also
+     files are considered invalid. Unfortunately this also
      matches files with filenames that actually end in " (deleted)",
      but there is not much to do about this. */
   if (g_str_has_suffix (path_buffer, " (deleted)"))
-    return NULL;
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_FILENAME,
+                   "Cannot share deleted file: %s", path_buffer);
+      return NULL;
+    }
 
   /* remap from sandbox to host if needed */
   return xdp_app_info_remap_path (app_info, path_buffer);
@@ -899,6 +1096,9 @@ xdp_get_alternate_document_path (const char *path, const char *app_id)
 {
   int len;
 
+  if (g_str_equal (app_id, ""))
+    return NULL;
+
   /* If we don't know where the document portal is mounted, then there
    * is no alternate path */
   if (documents_mountpoint == NULL)
@@ -913,45 +1113,115 @@ xdp_get_alternate_document_path (const char *path, const char *app_id)
   return g_strconcat (documents_mountpoint, "/by-app/", app_id, &path[len], NULL);
 }
 
+static gboolean
+check_same_file (const char *path,
+                 struct stat *expected_st_buf,
+                 GError **error)
+{
+  struct stat real_st_buf;
+  int saved_errno;
+
+  if (stat (path, &real_st_buf) < 0)
+    {
+      saved_errno = errno;
+      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (saved_errno),
+                   "stat %s: %s", path, g_strerror (saved_errno));
+      return FALSE;
+    }
+
+  if (expected_st_buf->st_dev != real_st_buf.st_dev ||
+      expected_st_buf->st_ino != real_st_buf.st_ino)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                   "\"%s\" identity (%ju,%ju) does not match expected (%ju,%ju)",
+                   path,
+                   (uintmax_t) expected_st_buf->st_dev,
+                   (uintmax_t) expected_st_buf->st_ino,
+                   (uintmax_t) real_st_buf.st_dev,
+                   (uintmax_t) real_st_buf.st_ino);
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
 char *
 xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
                               int fd,
                               int require_st_mode,
                               struct stat *st_buf,
-                              gboolean *writable_out)
+                              gboolean *writable_out,
+                              GError **error)
 {
   g_autofree char *proc_path = NULL;
   int fd_flags;
   struct stat st_buf_store;
-  struct stat real_st_buf;
   gboolean writable = FALSE;
   g_autofree char *path = NULL;
+  int saved_errno;
 
   if (st_buf == NULL)
     st_buf = &st_buf_store;
 
   if (fd == -1)
-    return NULL;
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                   "Invalid file descriptor");
+      return NULL;
+    }
 
   /* Must be able to get fd flags */
   fd_flags = fcntl (fd, F_GETFL);
   if (fd_flags == -1)
-    return NULL;
+    {
+      saved_errno = errno;
+      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (saved_errno),
+                   "Cannot get file descriptor flags (fcntl F_GETFL: %s)",
+                   g_strerror (saved_errno));
+      return NULL;
+    }
 
   /* Must be able to fstat */
   if (fstat (fd, st_buf) < 0)
-    return NULL;
+    {
+      saved_errno = errno;
+      g_set_error (error, G_IO_ERROR, g_io_error_from_errno (saved_errno),
+                   "Cannot get file information (fstat: %s)",
+                   g_strerror (saved_errno));
+      return NULL;
+    }
 
   /* Verify mode */
   if (require_st_mode != 0 &&
       (st_buf->st_mode & S_IFMT) != require_st_mode)
-    return NULL;
+    {
+      switch (require_st_mode)
+        {
+          case S_IFDIR:
+            g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_DIRECTORY,
+                         "File type 0o%o is not a directory",
+                         (st_buf->st_mode & S_IFMT));
+            return NULL;
+
+          case S_IFREG:
+            g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_REGULAR_FILE,
+                         "File type 0o%o is not a regular file",
+                         (st_buf->st_mode & S_IFMT));
+            return NULL;
+
+          default:
+            g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                         "File type 0o%o does not match expected 0o%o",
+                         (st_buf->st_mode & S_IFMT), require_st_mode);
+            return NULL;
+        }
+    }
 
   proc_path = g_strdup_printf ("/proc/self/fd/%d", fd);
 
   /* Must be able to read valid path from /proc/self/fd */
   /* This is an absolute and (at least at open time) symlink-expanded path */
-  path = verify_proc_self_fd (app_info, proc_path);
+  path = verify_proc_self_fd (app_info, proc_path, error);
   if (path == NULL)
     return NULL;
 
@@ -972,10 +1242,19 @@ xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
 
       /* Must not be O_NOFOLLOW (because we want the target file) */
       if ((fd_flags & O_NOFOLLOW) == O_NOFOLLOW)
-        return NULL;
+        {
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                       "O_PATH fd was opened O_NOFOLLOW");
+          return NULL;
+        }
 
       if (!xdp_app_info_supports_opath (app_info))
-        return NULL;
+        {
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
+                       "App \"%s\" of type %d does not support O_PATH fd passing",
+                       app_info->id, app_info->kind);
+          return NULL;
+        }
 
       read_access_mode = R_OK;
       if (S_ISDIR (st_buf->st_mode))
@@ -984,7 +1263,12 @@ xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
       /* Must be able to access the path via the sandbox supplied O_PATH fd,
          which applies the sandbox side mount options (like readonly). */
       if (access (proc_path, read_access_mode) != 0)
-        return NULL;
+        {
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
+                       "\"%s\" not available for read access via \"%s\"",
+                       path, proc_path);
+          return NULL;
+        }
 
       if (xdp_app_info_is_host (app_info) || access (proc_path, W_OK) == 0)
         writable = TRUE;
@@ -1006,9 +1290,7 @@ xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
     }
 
   /* Verify that this is the same file as the app opened */
-  if (stat (path, &real_st_buf) < 0 ||
-      st_buf->st_dev != real_st_buf.st_dev ||
-      st_buf->st_ino != real_st_buf.st_ino)
+  if (!check_same_file (path, st_buf, error))
     {
       /* If the path is provided by the document portal, the inode
          number will not match, due to only a subtree being mounted in
@@ -1021,10 +1303,12 @@ xdp_app_info_get_path_for_fd (XdpAppInfo *app_info,
       g_autofree char *alt_path = NULL;
       alt_path = xdp_get_alternate_document_path (path, xdp_app_info_get_id (app_info));
 
-      if (alt_path == NULL ||
-          stat (alt_path, &real_st_buf) < 0 ||
-          st_buf->st_dev != real_st_buf.st_dev ||
-          st_buf->st_ino != real_st_buf.st_ino)
+      if (alt_path == NULL)
+        return NULL;
+
+      g_clear_error (error);
+
+      if (!check_same_file (alt_path, st_buf, error))
         return NULL;
     }
 
@@ -1100,22 +1384,6 @@ xdp_is_valid_app_id (const char *string)
 }
 
 
-static gboolean
-needs_quoting (const char *arg)
-{
-  while (*arg != 0)
-    {
-      char c = *arg;
-      if (!g_ascii_isalnum (c) &&
-          !(c == '-' || c == '/' || c == '~' ||
-            c == ':' || c == '.' || c == '_' ||
-            c == '='))
-        return TRUE;
-      arg++;
-    }
-  return FALSE;
-}
-
 char *
 xdp_quote_argv (const char *argv[])
 {
@@ -1566,7 +1834,7 @@ map_pids (DIR     *proc,
       pid_t inside = 0;
       uid_t uid = 0;
       guint idx;
-      ino_t ns;
+      ino_t ns = 0;
       int r;
 
       if (de->d_type != DT_DIR)
@@ -1808,13 +2076,15 @@ xdg_app_info_ensure_pidns (XdpAppInfo  *app_info,
                            GError     **error)
 {
   g_autoptr(JsonNode) root = NULL;
-  xdp_lockguard GMutex *guard = NULL;
+  g_autoptr(GMutexLocker) guard = NULL;
   xdp_autofd int fd = -1;
   pid_t pid;
   ino_t ns;
   int r;
 
-  guard = xdp_auto_lock_helper (app_info->u.flatpak.pidns_lock);
+  g_assert (app_info->kind == XDP_APP_INFO_KIND_FLATPAK);
+
+  guard = g_mutex_locker_new (&(app_info->u.flatpak.pidns_lock));
 
   if (app_info->u.flatpak.pidns_id != 0)
     return TRUE;
@@ -1934,3 +2204,185 @@ xdg_app_info_pidfds_to_pids (XdpAppInfo  *app_info,
 
   return ok;
 }
+
+void
+cleanup_temp_file (void *p)
+{
+  void **pp = (void **)p;
+
+  if (*pp)
+    remove (*pp);
+  g_free (*pp);
+}
+
+#define ICON_VALIDATOR_GROUP "Icon Validator"
+
+gboolean
+xdp_validate_serialized_icon (GVariant  *v,
+                              gboolean   bytes_only,
+                              char     **out_format,
+                              char     **out_size)
+{
+  g_autoptr(GIcon) icon = NULL;
+  GBytes *bytes;
+  __attribute__((cleanup(cleanup_temp_file))) char *name = NULL;
+  xdp_autofd int fd = -1;
+  g_autoptr(GOutputStream) stream = NULL;
+  int status;
+  g_autofree char *format = NULL;
+  g_autofree char *stdoutlog = NULL;
+  g_autofree char *stderrlog = NULL;
+  g_autoptr(GError) error = NULL;
+  const char *icon_validator = LIBEXECDIR "/xdg-desktop-portal-validate-icon";
+  const char *args[6];
+  /* same allowed formats as Flatpak */
+  const char *allowed_icon_formats[] = { "png", "jpeg", "svg", NULL };
+  int size;
+  const char *MAX_ICON_SIZE = "512";
+  g_auto(GStrv) parts = NULL;
+  gconstpointer bytes_data;
+  gsize bytes_len;
+  g_autoptr(GKeyFile) key_file = NULL;
+
+  icon = g_icon_deserialize (v);
+  if (!icon)
+    {
+      g_warning ("Icon deserialization failed");
+      return FALSE;
+    }
+
+  if (!bytes_only && G_IS_THEMED_ICON (icon))
+    {
+      g_autofree char *a = g_strjoinv (" ", (char **)g_themed_icon_get_names (G_THEMED_ICON (icon)));
+      g_debug ("Icon validation: themed icon (%s) is ok", a);
+      return TRUE;
+    }
+
+  if (!G_IS_BYTES_ICON (icon))
+    {
+      g_warning ("Unexpected icon type: %s", G_OBJECT_TYPE_NAME (icon));
+      return FALSE;
+    }
+
+  if (!g_file_test (icon_validator, G_FILE_TEST_EXISTS))
+    {
+      g_warning ("Icon validation: %s not found, rejecting icon by default.", icon_validator);
+      return FALSE;
+    }
+
+  bytes = g_bytes_icon_get_bytes (G_BYTES_ICON (icon));
+  fd = g_file_open_tmp ("iconXXXXXX", &name, &error);
+  if (fd == -1)
+    {
+      g_warning ("Icon validation: %s", error->message);
+      return FALSE;
+    }
+
+  stream = g_unix_output_stream_new (fd, FALSE);
+
+  /* Use write_all() instead of write_bytes() so we don't have to worry about
+   * partial writes (https://gitlab.gnome.org/GNOME/glib/-/issues/570).
+   */
+  bytes_data = g_bytes_get_data (bytes, &bytes_len);
+  if (!g_output_stream_write_all (stream, bytes_data, bytes_len, NULL, NULL, &error))
+    {
+      g_warning ("Icon validation: %s", error->message);
+      return FALSE;
+    }
+
+  if (!g_output_stream_close (stream, NULL, &error))
+    {
+      g_warning ("Icon validation: %s", error->message);
+      return FALSE;
+    }
+
+  args[0] = icon_validator;
+  args[1] = "--sandbox";
+  args[2] = MAX_ICON_SIZE;
+  args[3] = MAX_ICON_SIZE;
+  args[4] = name;
+  args[5] = NULL;
+
+  if (!g_spawn_sync (NULL, (char **)args, NULL, 0, NULL, NULL, &stdoutlog, &stderrlog, &status, &error))
+    {
+      g_warning ("Icon validation: %s", error->message);
+      g_warning ("stderr:\n%s\n", stderrlog);
+      return FALSE;
+    }
+
+  if (!g_spawn_check_exit_status (status, &error))
+    {
+      g_warning ("Icon validation: %s", error->message);
+      return FALSE;
+    }
+
+  key_file = g_key_file_new ();
+  if (!g_key_file_load_from_data (key_file, stdoutlog, -1, G_KEY_FILE_NONE, &error))
+    {
+      g_warning ("Icon validation: %s", error->message);
+      return FALSE;
+    }
+  if (!(format = g_key_file_get_string (key_file, ICON_VALIDATOR_GROUP, "format", &error)) ||
+      !g_strv_contains (allowed_icon_formats, format))
+    {
+      g_warning ("Icon validation: %s", error ? error->message : "not allowed format");
+      return FALSE;
+    }
+  if (!(size = g_key_file_get_integer (key_file, ICON_VALIDATOR_GROUP, "width", &error)))
+    {
+      g_warning ("Icon validation: %s", error->message);
+      return FALSE;
+    }
+
+  if (out_format)
+    *out_format = g_steal_pointer (&format);
+  if (out_size)
+    *out_size = g_strdup_printf ("%d", size);
+
+  return TRUE;
+}
+
+#if !GLIB_CHECK_VERSION (2, 68, 0)
+/* All this code is backported directly from glib */
+guint
+g_string_replace (GString     *string,
+                  const gchar *find,
+                  const gchar *replace,
+                  guint        limit)
+{
+  gsize f_len, r_len, pos;
+  gchar *cur, *next;
+  guint n = 0;
+
+  g_return_val_if_fail (string != NULL, 0);
+  g_return_val_if_fail (find != NULL, 0);
+  g_return_val_if_fail (replace != NULL, 0);
+
+  f_len = strlen (find);
+  r_len = strlen (replace);
+  cur = string->str;
+
+  while ((next = strstr (cur, find)) != NULL)
+    {
+      pos = next - string->str;
+      g_string_erase (string, pos, f_len);
+      g_string_insert (string, pos, replace);
+      cur = string->str + pos + r_len;
+      n++;
+      /* Only match the empty string once at any given position, to
+       * avoid infinite loops */
+      if (f_len == 0)
+        {
+          if (cur[0] == '\0')
+            break;
+          else
+            cur++;
+        }
+      if (n == limit)
+        break;
+    }
+
+  return n;
+}
+
+#endif /* GLIB_CHECK_VERSION (2, 28, 0) */
